"""
.. module:: orgasm.tango
   :platform: Unix
   :synopsis: The :py:mod:`~orgasm.tango` package contains a set of functions useful to manage assembling structure.

The :py:mod:`orgasm.tango` package
==================================

The :py:mod:`~orgasm.tango` package contains a set of functions useful to manage assembling structure.

.. moduleauthor:: Eric Coissac <eric.coissac@inria.fr>

:author:  Eric Coissac
:contact: eric.coissac@inria.fr


.. warning::

    The :py:mod:`~orgasm.tango` package functions aims to be integrated 
    into other packages as standalone functions of class methods.
"""
import sys
from collections import Counter
from itertools import imap
from orgasm.multialign import multiAlignReads, consensus

def tango(self,seeds,
          minread=10,    minratio=0.1, 
          mincov=1,      minoverlap=40,
          lowfilter=True,maxjump=0,restrict=None,
          cycle=1):
    '''
    the :py:func:`~organsm.tango.tango` function is the main assembling function. It extends selected
    seeds to produce the assembled sequence.
    
    :param seeds: a :py:class:`list` describing the extensions points. Each element of the list corresponds to 
                  a seed. A seed is described by a :py:class:`tuple` of two elements. 
                  
                  The first element is an integer value corresponding to the ``id`` of a read. if the ``id`` is greater than 0
                  the seed will be extended only if it is not already included in the assembling graph.
                  If the ``id`` is specified as a negative number, the seed will be included in the procedure
                  extension whatever if it is already member or not of the assembling graph.
                  
                  The second element of the :py:class:`tuple` is itself a :py:class:`tuple` contenaing or the 
                  :py:class:`None` value of a :py:class:`str` elements used to annotate the read.
                  
                  .. code-block :: python
                  
                      [(12347, ('matK',)),(-37247,(None,)]
                  
                  In this example the structure describe two seeds. The first one corresponds to the
                  read 12347 annotated as belonging to the *matK* gene. The second corresponds to the
                  read 37247, without annotation. The process of extension will be run for the first seed
                  only if it is not already member of the assembling graph and always run on the second seed.
                  
    :type seeds: a :py:class:`list` of :py:class:`tuple`.
                     
    :param minread: Minimum count of reads having to be taken into account to extend an assembling path.
                    If this minimum count is not reach the extension of this path is stopped
    :type minread: :py:class:`int`
    
    :param minratio: A branch is added to the current path if the corresponding symbole **A**, **C**, **G** or **T**
                     represents at least a fraction of ``minratio``of the observed reads.
    :type minratio: :py:class:`float`
    
    :param mincov: To be considered, a read must be observed at least ``mincov`` in the total data set.
    :type mincov: :py:class:`int`
    
    :param minoverlap: Only reads overlaping the current extension point by ``minoverlap`` base pairs are
                       considered for the extension process. 
    :type minoverlap: :py:class:`int`
    
    :param lowfilter:
    :type lowfilter: :py:class:`bool`
    :param maxjump:
    :type maxjump: :py:class:`int`
    :param restrict:
    :type restrict: :py:class:`set` of :py:class:`int` values
    '''
        
    # a direct pointer to the read index
    index = self.index
    
    # a direct pointer to the graph structure
    graph = self.graph
    
    # a direct access to the read length
    readLength = index.getReadSize()
    readmax    = len(index) + 1
    
    # extension cycle counter for the progress status monitoring
    icycle=0
    
    # count of missing reads in the graph (aka fake reads)
    fake=0
    
    componantes=0
    
    # list of the stack size during the last 1000 cycles 
    lastcycle=[]
    
    # count of pair end jump on an extension branch
    maxanchor = 0
    
    while(seeds):
        
        #time.sleep(1)
        icycle+=1
        
        # Counter used to check how many seed will be added 
        # to the stack during the cycle
        addedseed=0

        # We pop one seed from the stack
        
        readid,(gene,) = seeds.pop()
        
        # We test if the read poped out from the stack
        # is just a jump mark
        if readid==0:
            maxanchor-=1
            continue
        
        if readid < 0 or readid not in graph:
            
            if readid < 0:
                readid=-readid
            #print "\n==>",readid,str(gene)
            #
            # We add the read into the graph
            #
            
            lastcycle.append(len(seeds))
            if len(lastcycle) > 1000:
                lastcycle=lastcycle[-1000:]         
            print >>sys.stderr,b"Cycle : %8d  (%d nodes / %4.1f%% fake) Waiting points : %8d / %6.2f  Gene: %s \r" % (icycle,len(graph),float(fake)/(len(graph)+1)*200.,len(seeds),float(sum(lastcycle))/len(lastcycle),gene),
            #print >>sys.stderr,b"Cycle : %8d  (%d nodes) Waiting points : %8d    " % (cycle,len(graph),len(seeds))
    
                
            #
            # extends the read on the right side
            #
                
            rextensions = index.checkedExtensions(index.getRead(readid,1,readLength-1), 
                                                 minread, minratio, 
                                                 mincov,  minoverlap, 
                                                 1,
                                                 restrict=restrict)
            
            #
            # Build a bijective relationship for the right extensions
            #
            
            for k in rextensions.keys():
                backe = index.checkedExtensions(index.getRead(-rextensions[k][1][0],1,readLength-1),
                                                minread, minratio, 
                                                mincov,  minoverlap, 
                                                1,
                                                restrict=restrict)
                back = set(reduce(lambda p,q:p+q,[backe[y][1] for y in backe],[]))
                
                if -readid not in back:
                    #print "Remove R extension : ",k
                    del rextensions[k]
            
    
            lextensions = index.checkedExtensions(index.getRead(-readid,1,readLength-1), 
                                                 minread, minratio, 
                                                 mincov,  minoverlap, 
                                                 1,
                                                 restrict=restrict)
     
            #
            # Build a bijective relationship for the left extensions
            #
    
            for k in lextensions.keys():
                backe = index.checkedExtensions(index.getRead(-lextensions[k][1][0],1,readLength-1), 
                                                 minread, minratio, 
                                                 mincov,  minoverlap, 
                                                 1,
                                                 restrict=restrict)
                
                back = set(reduce(lambda p,q:p+q,[backe[y][1] for y in backe],[]))
                
                if readid not in back:
                    #print "Remove L extension : ",k
                    del lextensions[k]
                    
    #        print "Rext = ",rextensions
    #        print "Lext = ",lextensions
    
            if rextensions or lextensions:
                node  = graph.addNode(readid)
                node['gene']=gene
                if 'cycle' not in node:
                    node['cycle']=cycle
                if readid < readmax:
                    node['fake5']=0
                    node['fake3']=0
                else:
                    fake+=1
                    node['graphics']={'fill':"#00FF00"}
    
    #        print "R",len(seq)
    
            for ex in rextensions:
                data = rextensions[ex]
                coverage = data[0]
    #            print ex,data
                
                nodeindexE,c,s = index.getIds(data[1][0])
                                    
                if nodeindexE in graph:
                    if data[1][0] not in s:
                        nodeindexE=-nodeindexE
                    if readid!=-nodeindexE:
                        edges = graph.addEdge(readid,nodeindexE)
                        edges[1][b'coverage']=coverage
                        edges[2][b'coverage']=coverage
                        edges[1][b'label']="%s (%d)" % (edges[1][b'ext'],coverage)
                        edges[2][b'label']="%s (%d)" % (edges[2][b'ext'],coverage)
                        edges[1][b'type']=b"next read"
                        edges[2][b'type']=b"next read"
                    else:
                        print >>sys.stderr,"\nWARNING : self loop on %d" % nodeindexE
      #              print "linked"
                else:
      #              print "not linked"
                    seeds.append((nodeindexE,(None,)))
                    addedseed+=1
                                
            #
            # extends the read on the left side
            # which is equivalent to extend on the right side
            # the reverse complement of the read
            #
            
      #      print "L",len(cseq)
            
     
            for ex in lextensions:
                data = lextensions[ex]
                coverage = data[0]
    #            print ex,data
    
                nodeindexE,c,s = index.getIds(data[1][0])
                                        
                if nodeindexE in graph:
                    if data[1][0] not in s:
                        nodeindexE=-nodeindexE
                    edges = graph.addEdge(-readid,nodeindexE)
                    edges[1][b'coverage']=coverage
                    edges[2][b'coverage']=coverage
                    edges[1][b'label']="%s (%d)" % (edges[1][b'ext'],coverage)
                    edges[2][b'label']="%s (%d)" % (edges[2][b'ext'],coverage)
                    edges[1][b'type']=b"next read"
                    edges[2][b'type']=b"next read"
      #              print "linked"
                else:
      #              print "not linked"
                    seeds.append((nodeindexE,(None,)))
                    addedseed+=1
    
            if addedseed==0 and maxjump > 0 and restrict is None:
                if maxanchor < maxjump and readid < readmax and readid in graph:
                    maxanchor+=1
                    seeds.append((0,(None,)))
                    anchors = map(lambda id:index.getIds(index.getPairedRead(id))[0],index.getIds(readid)[2])
                    
                    print >>sys.stderr,"\n\nJumpGap on read %d\n" % readid
                    for a in anchors:                        
                        if a not in graph:
                            seeds.append((-a,(None,)))
                        
    return graph

def cutLowCoverage(self,mincov,terminal=True):
    '''
    Remove sequences in the assembling graph with a coverage below ``mincov``.
    
    .. code-block :: ipython
    
        In [159]: asm = Assembler(r)

        In [160]: s = matchtoseed(m,r)

        In [161]: a = tango(asm,s,mincov=1,minread=10,minoverlap=30,maxjump=0,cycle=1)
    
        In [162]: asm.cleanDeadBranches(maxlength=10)

        Remaining edges : 424216 node : 423896
        Out[162]: 34821
        In [162]: cutLowCoverage(asm,10,terminal=False)

    :param mincov: coverage threshold 
    :type mincov: :py:class:`int`
    :param terminal: if set to ``True`` only terminal edges are removed from the assembling graph
    :type terminal: :py:class:`bool`
    :return: the count of deleted node
    :rtype: :py:class:`int`
    
    :seealso: :py:meth:`~orgasm.assambler.Assembler.cleanDeadBranches`
    '''
    def isTerminal(g,n):
        return len(list(g.parentIterator(n)))==0 or len(list(g.neighbourIterator(n)))==0
    
    def endnodeset(g):
        return set(g.nodeIterator(predicate = lambda n : (len(list(g.neighbourIterator(n)))==0)))
    
    def startnodeset(g):
        return set(g.nodeIterator(predicate = lambda n : (len(list(g.parentIterator(n)))==0)))
    
    
    if terminal:
        tstates=[True]
    else:
        tstates=[True,False]
    ilength=len(self)
    cg = self.compactAssembling(verbose=False)
    index = self.index
    readSize = index.getReadSize()
    for terminal in tstates:
        print >>sys.stderr
        if terminal:
            print >>sys.stderr,"Deleting terminal branches"
        else:
            print >>sys.stderr,"Deleting internal branches"
        extremities = endnodeset(cg) | startnodeset(cg)
        go = True
        while go:
            go = False
            stems = [x for x in cg.edgeIterator() 
                     if not terminal or (isTerminal(cg, x[0]) or isTerminal(cg, x[1]))]
            if stems:
                stems.sort(key=lambda i:cg.getEdgeAttr(*i)[b'weight'],reverse=True)
                lightest = stems.pop()
                lattr = cg.getEdgeAttr(*lightest)
#                 print >>sys.stderr
#                 print >>sys.stderr,lattr['stemid'],len(stems),lightest[0],lightest[1],lattr['weight']
        #         print >>sys.stderr,lattr['weight'] < mincov
        #         print >>sys.stderr,isTerminal(cg, lightest[0]), isTerminal(cg, lightest[1])
        #         print >>sys.stderr,lattr['weight'] < mincov and ((not terminal) or isTerminal(cg, lightest[0]) or isTerminal(cg, lightest[1]))
        #         print >>sys.stderr
                if lattr['weight'] < mincov: 
                    if stems:
                        go=True
                    for n in lattr['path'][1:-1]:
                        if n in self.graph:
                            del self.graph[n]
                    if lightest[0] in extremities and lightest[0] in self.graph:
                        del self.graph[lightest[0]]
                    if lightest[1] in extremities and lightest[1] in self.graph:
                        del self.graph[lightest[1]]
                        
                    print >>sys.stderr,"Remaining edges : %d node : %d\r" % (self.graph.edgeCount(),len(self)),
        
                    cg.deleteEdge(*lightest)
                    tojoin=[]
                    if lightest[0] in extremities:
                        del cg[lightest[0]]
                    else:
                        tojoin.append(lightest[0])
                    if lightest[1] in extremities:
                        del cg[lightest[1]]
                    else:
                        tojoin.append(lightest[1])
#                    print >>sys.stderr,lightest[0] in extremities,lightest[1] in extremities,tojoin
                    for c in tojoin:
                        begin = list(cg.parentIterator(c))
                        end   = list(cg.neighbourIterator(c))
                        if len(begin)==1 and len(end)==1:
                            begin = begin[0]
                            end = end[0]
                            e1s = list(cg.edgeIterator(edgePredicate = lambda e:e[0]==begin and e[1]==c))
                            e2s = list(cg.edgeIterator(edgePredicate = lambda e:e[0]==c and e[1]==end))
                            if len(e1s)==1 and len(e2s)==1:
                                e1 = e1s[0]
                                e2 = e2s[0]
                                attr1 = cg.getEdgeAttr(*e1)
                                
                                attr2 = cg.getEdgeAttr(*e2)
                                lentot= attr1['length'] + attr2['length']
                                weight= float(attr1['weight'] * attr1['length'] + attr2['weight'] * attr2['length'])/lentot
#                                     print >>sys.stderr
#                                     print >>sys.stderr,"(",attr1['weight'],"*", attr1['length'], '+',attr2['weight'],"*",attr2['length'],")/",lentot,"=",weight
#                                     time.sleep(1)
                                sequence=attr1['sequence'] + attr2['sequence']
                                stemid=lattr['stemid']
                                path=attr1['path'] + attr2['path'][1:]
                                graphics={b'width':int(weight/readSize)+1,
                                  b'arrow':b'last'}
                        
                                if lentot > 10:
                                    label=b"%d : %s->(%d)->%s  [%d]" % (stemid,sequence[0:5],lentot,sequence[-5:],weight)
                                else:
                                    label=b"%d : %s->(%d)  [%d]" % (stemid,sequence,lentot,weight)
                                
 #                               print >>sys.stderr,"Adding merged edge -->",stemid,begin,end,weight
                                attr = cg.addEdge(begin,end)
                                attr[b'sequence']=sequence
                                attr[b'stemid']=stemid
                                attr[b'path']=path
                                attr[b'length']=lentot
                                attr[b'weight']= weight
                                attr[b'graphics']=graphics
                                attr[b'label']=label

                                cg.deleteEdge(*e2)                                        
                                cg.deleteEdge(*e1)  
                                del cg[c]
                                      
                                
    print >>sys.stderr      
    return ilength - len(self)
  
def matchtoseed(matches,index):
    k = matches.keys()
    s=[]
    for x in k:
        s.extend((index.getIds(y[0])[0],(x,)) for y in matches[x])  

    return s

def scaffold(self,assgraph,minlink=5,back=200,addConnectedLink=False):
    '''
    Add relationships between edges of the assembling graph related to the
    par ended links.
    
    :param assgraph: The compact assembling graph as produced by the
                     :py:meth:`~orgasm.assembler.Assembler.compactAssembling` method
    :type assgraph:  :py:class:`~orgasm.graph.DiGraphMultiEdge`
    :param minlink:  the minimum count of pair ended link to consider 
                     for asserting the relationship
    :type minlink:   :py:class:`int`
    :param back:     How many base pairs must be considered at the end of each edge
    :type back:      :py:class:`int`
    :param addConnectedLink: add to the assembling graph green edges for each directly 
                             connected edge pair representing the pair ended links 
                             asserting the connection.
    :type addConnectedLink: :py:class:`bool`
     '''
    
    if addConnectedLink:
        # Fake function just testing the stemid attribute
        def isInitial(n):
            return  'stemid' in assgraph.getEdgeAttr(*n) 
        def isTerminal(n):
            return  'stemid' in assgraph.getEdgeAttr(*n) 
    else:
        def isInitial(n):
            return len(list(assgraph.parentIterator(n[0],edgePredicate=lambda e: 'stemid' in assgraph.getEdgeAttr(*e))))==0 
        def isTerminal(n):
            return len(list(assgraph.neighbourIterator(n[1],edgePredicate=lambda e: 'stemid' in assgraph.getEdgeAttr(*e))))==0

        
    ei = [i for i in assgraph.edgeIterator(edgePredicate=isInitial)]
    et = [i for i in assgraph.edgeIterator(edgePredicate=isTerminal)]
    
    eiid = [assgraph.getEdgeAttr(*i)['stemid'] for i in ei]
    etid = [assgraph.getEdgeAttr(*i)['stemid'] for i in et]

    epi = [set(assgraph.getEdgeAttr(*i)['path'][0:back]) for i in ei]
    ept = [set(assgraph.getEdgeAttr(*i)['path'][-back:]) for i in et]
    
    exi = [getPairedRead(self,assgraph,i,back,end=False) for i in eiid]
    ext = [getPairedRead(self,assgraph,i,back,end=True) for i in etid]
    
    nei = len(ei)
    net = len(et)

    s=[]
    for e1 in xrange(net):
        for e2 in xrange(nei):
            connected = et[e1][1]==ei[e2][0]
            linkedby  = len(ext[e1] & exi[e2] | ext[e1] & epi[e2] | ept[e1] & exi[e2]) 
                       
            if connected and addConnectedLink:
                if linkedby >= minlink:
                    s.append((et[e1][1],ei[e2][0],linkedby,etid[e1], eiid[e2],"#00FF00"))
                    print >>sys.stderr,"%d -> %d connection asserted by %d pair ended links" % (etid[e1], eiid[e2],linkedby)
                else:
                    print >>sys.stderr,"%d -> %d connection not asserted by pair ended link" % (etid[e1], eiid[e2])
                    
            if not connected and linkedby >= minlink:
                s.append((et[e1][1],ei[e2][0],linkedby,etid[e1], eiid[e2],"#FF0000"))
                print >>sys.stderr,"%d -> %d scaffolded by %d pair ended links" % (etid[e1], eiid[e2],linkedby)
    
    for x,y,z,s1,s2,color in s:
        attr = assgraph.addEdge(x,y)
        attr[b'label']=b"Gap (%d)  %d -> %d" % (z,s1,s2)
        attr[b'graphics']={b'width':z / 10.,
                           b'arrow':b'last',
                           b'fill':color
                          }


def fillGaps(self,minlink=5,
                  back=200,
                  kmer=12,
                  smin=40,
                  delta=0,
                  cmincov=5,
                  minread=20,
                  minratio=0.1, 
                  emincov=1,
                  maxlength=10,
                  gmincov=1,
                  minoverlap=60,
                  lowfilter=True,maxjump=0):
    def isInitial(n):
        return len(list(assgraph.parentIterator(n[0],edgePredicate=lambda e: 'stemid' in assgraph.getEdgeAttr(*e))))==0 
    def isTerminal(n):
        return len(list(assgraph.neighbourIterator(n[1],edgePredicate=lambda e: 'stemid' in assgraph.getEdgeAttr(*e))))==0

    assgraph = self.compactAssembling(verbose=False)
    index = self.index
    ei = [i for i in assgraph.edgeIterator(edgePredicate=isInitial)]
    et = [i for i in assgraph.edgeIterator(edgePredicate=isTerminal)]
    
    eiid = [assgraph.getEdgeAttr(*i)['stemid'] for i in ei]
    etid = [assgraph.getEdgeAttr(*i)['stemid'] for i in et]

    epi = [set(assgraph.getEdgeAttr(*i)['path'][0:back]) for i in ei]
    ept = [set(assgraph.getEdgeAttr(*i)['path'][-back:]) for i in et]
    
    exi = [getPairedRead(self,assgraph,i,back,end=False) for i in eiid]
    ext = [getPairedRead(self,assgraph,i,back,end=True) for i in etid]
    
    nei = len(ei)
    net = len(et)
    s=[]
    maxcycle=max(self.graph.getNodeAttr(i)['cycle'] for i in self.graph)
    lassemb = len(self)
    cycle = maxcycle
    for e1 in xrange(net):
        free=True
        for e2 in xrange(nei):
            connected = et[e1][1]==ei[e2][0]
            if not connected:
                linkedby  = len(ext[e1] & exi[e2] | ext[e1] & epi[e2] | ept[e1] & exi[e2]) 
            
                if linkedby >= minlink and etid[e1] <= eiid[e2]:
                    free=False
                    print >>sys.stderr,"Linking Stems %d -> %d" % (etid[e1],eiid[e2])
                    ex = ext[e1] | exi[e2]
                    nreads = len(ex)
                    print >>sys.stderr,"--> %d | %d = %d reads to align" % (len(ext[e1]),len(exi[e2]),nreads)
                    if nreads > 10:
                        ali= multiAlignReads(ex,index,kmer,smin,delta)
                        print >>sys.stderr 
    
                        goodali = [i for i in ali if len(i) >= nreads/4]
                        print >>sys.stderr,"--> %d consensus to add" % len(goodali)
                        for a in goodali:
                            cycle+=1
                            print >>sys.stderr,"       added on cycle %d" % cycle
                            c = consensus(a,index,cmincov)
                            s = insertFragment(self,c,cycle=cycle)
                            a = tango(self,s,mincov=emincov,
                                             minread=minread,
                                             minoverlap=minoverlap,
                                             maxjump=maxjump,cycle=cycle)
                            print >>sys.stderr 
            else:
                free=False
        if free:
            print >>sys.stderr,"Extending Stems %d" % (etid[e1])
            ex = ext[e1]
            nreads = len(ex)
            print >>sys.stderr,"--> %d reads to align" % (nreads)
            if nreads > 10:
                ali= multiAlignReads(ex,index,kmer,smin,delta)
                print >>sys.stderr 
                goodali = [i for i in ali if len(i) >= nreads/4]
                print >>sys.stderr,"--> %d consensus to add" % len(goodali)
                for a in goodali:
                    c = consensus(a,index,cmincov)
                    if c:
                        cycle+=1
                        print >>sys.stderr,"       added on cycle %d" % cycle
                        s = insertFragment(self,c,cycle=cycle)
                        a = tango(self,s,mincov=emincov,
                                         minread=minread,
                                         minoverlap=minoverlap,
                                         maxjump=maxjump,cycle=cycle)
                    print >>sys.stderr 

    self.cleanDeadBranches(maxlength=maxlength)
    cutLowCoverage(self,gmincov,terminal=False)
    
    #newnodes=sum(1 for i in self.graph if self.graph.getNodeAttr(i)['cycle'] > maxcycle)
    
    newnodes = len(self) - lassemb
    
    print >>sys.stderr
    print >>sys.stderr,"###################################"
    print >>sys.stderr,"#"
    print >>sys.stderr,"# Added nodes : %d" % newnodes
    print >>sys.stderr,"#"
    print >>sys.stderr,"###################################"
    print >>sys.stderr
    
    return newnodes

                        


        
def insertFragment(self,seq,cycle=1):
    index = self.index
    rsize = index.getReadSize()
    readmax=len(index)+1
    seeds = set()
    ireadidE=None
    
    if len(seq) >= rsize:
        graph = self.graph
        probe = seq[0:rsize]
        readid = index.getReadIds(probe)
        
        for i in xrange(1,len(seq)-rsize+1):
            coverage = readid[1]
            ireadid   = readid[0]
            seeds.add(ireadid)
            
            if ireadid not in graph:
                node  = graph.addNode(ireadid)
                if 'cycle' not in node:
                    node['cycle']=cycle
                if ireadid < readmax:
                    node['fake5']=0
                    node['fake3']=0
                else:
                    node['graphics']={'fill':"#0000FF"}
                
            probe = seq[i:i+rsize]
            readidE = index.getReadIds(probe)
            coverage = readidE[1]
            ireadidE   = readidE[0]
            
            if ireadidE not in graph:
                node  = graph.addNode(ireadidE)
                if 'cycle' not in node:
                    node['cycle']=cycle
                if ireadidE < readmax:
                    node['fake5']=0
                    node['fake3']=0
                else:
                    node['graphics']={'fill':"#0000FF"}
                    
            
            if ireadid!=-ireadidE:
                edges = graph.addEdge(ireadid,ireadidE)
                edges[1][b'coverage']=coverage
                edges[2][b'coverage']=coverage
                edges[1][b'label']="%s (%d)" % (edges[1][b'ext'],coverage)
                edges[2][b'label']="%s (%d)" % (edges[2][b'ext'],coverage)
                edges[1][b'type']=b"next read"
                edges[2][b'type']=b"next read"
            else:
                print >>sys.stderr,"\nWARNING : self loop on %d" % readidE
        
            readid = readidE
        
        if ireadidE is not None:    
            seeds.add(ireadidE)
        
    return [(-abs(i),("Consensus",)) for i in seeds]
                

def getPairedRead(self,assgraph,stemid,back,end=True): 
    '''
    
    :param assgraph:
    :type assgraph:
    :param stemid:
    :type stemid:
    :param back:
    :type back:
    :param end:
    :type end:
    '''
    r = self.index 
    if not end:
        stemid=-stemid  
    stem = assgraph.edgeIterator(edgePredicate=lambda e:assgraph.getEdgeAttr(e[0],e[1],e[2])['stemid']==stemid).next()
    path=assgraph.getEdgeAttr(*stem)['path']
    path=[i for i in path[-back:] if i > 0]
    reads = [(i,) + r.getIds(i) for i in path if i <= len(r)]
    reads=[i[3] if i[0] in i[3] else set(-j for j in i[3]) for i in reads]
    paired=[(r.getPairedRead(j) for j in i) for i in reads]
    paired=[[(j,) + r.getIds(j) for j in i if j!=0] for i in paired]
    paired = [[j[1] if j[0] in j[3] else -j[1] for j in i] for i in paired]
    paired = set(reduce(lambda a,b: a+b,(i for i in paired),[]))
    if not end:
        paired = set(-i for i in paired)
    return paired

def pairEndedConnected(self,assgraph,edge1,edge2,back=200):
    '''
    Returns how many pair ended reads link two edges in a compact assembling graph
    
    :param assgraph: The compact assembling graph as produced by the
                     :py:meth:`~orgasm.assembler.Assembler.compactAssembling` method
    :type assgraph:  :py:class:`~orgasm.graph.DiGraphMultiEdge`
    :param edge1:    The ``stemid`` of the first edge
    :type edge1:     :py:class:`int`
    :param edge2:    The ``stemid`` of the second edge
    :type edge2:     :py:class:`int`
    :param back:     How many base pairs must be considered at the end of each edge
    :type back:      :py:class:`int`
    
    :return:         The count of pair ended reads linking both the edges
    :rtype:          :py:class:`int`
    '''
    
    p1 = getPairedRead(self,assgraph,edge1,back,end=True)
    
    s2 = assgraph.edgeIterator(edgePredicate=lambda e:assgraph.getEdgeAttr(e[0],e[1],e[2])['stemid']==edge2).next()
    p2 = set(assgraph.getEdgeAttr(*s2)['path'][0:back])
    
    p3 = getPairedRead(self,assgraph,edge2,back,end=False)
    
    s4 = assgraph.edgeIterator(edgePredicate=lambda e:assgraph.getEdgeAttr(e[0],e[1],e[2])['stemid']==edge1).next()
    p4 = set(assgraph.getEdgeAttr(*s4)['path'][-back:])

    n = len(p1 & p2 | p3 & p4 | p1 & p3)
    
    return n
    
def coverageEstimate(self):
    def weight(a,b):
        maxw=0
        for e in cg.edgeIterator(edgePredicate = lambda i:i[0]==a and i[1]==b):
            attr = cg.getEdgeAttr(*e)
            w = attr['length'] * attr['weight']
            if w > maxw:
                maxw=w        
        return maxw 
    
    def coverage(start):
        wp = cg.minpath(start,distance=weight,allowCycle=True)
#        print ">>>",start,wp
        maxpath = max(wp[0].values())
        maxend=[i for i in wp[0] if wp[0][i]==maxpath][0]
        i=maxend
        path=[i]
        if i == start and wp[0][i] > 0:
            i=wp[1][i]
            path.append(i)
        while i!=start:
            i=wp[1][i]
            path.append(i)
        return(maxpath,path)
    
    cg = self.compactAssembling(verbose=False)
    paths = [coverage(i) for i in cg]
#    print paths
    maxpath=0
    path=None
    for w,p in paths:
        if w > maxpath:
            maxpath=w
            path=p 
#    print path        
    path.reverse()
    j = path[0]
    cumlength=0
    for i in path[1:]:
        maxw=0
        maxlength=0
        for e in cg.edgeIterator(edgePredicate = lambda k:k[0]==j and k[1]==i):
            attr = cg.getEdgeAttr(*e)
            w = attr['length'] * attr['weight']
#            print w , attr['length'] , attr['weight']
            if w > maxw:
                maxw=w
                maxlength=attr['length']       

        cumlength+=maxlength
        j=i
        
    return maxpath,cumlength,maxpath/float(cumlength)

def path2fasta(self,assgraph,path,identifier="contig",minlink=10,nlength=20,back=200):
    '''
    Convert a path in an compact assembling graph in a fasta formated sequences.
    
    :param assgraph: The compact assembling graph as produced by the
                     :py:meth:`~orgasm.assembler.Assembler.compactAssembling` method
    :type assgraph:  :py:class:`~orgasm.graph.DiGraphMultiEdge`
    :param path:     an ``iterable`` providing an ordered list of ``stemid`` indicating
                     the path to follow.  
    :type path:      an ``iterable`` over :py:class:`int`
    :param identifier: the identifier used in the header of the fasta formated sequence
    :type identifier:  :py:class:`bytes`
    :param minlink:  the minimum count of pair ended link to consider 
                     for asserting the relationship
    :type minlink:   :py:class:`int`
    :param nlength:  how many ``N`` must be added between two segment of sequences only connected
                     by pair ended links
    :type nlength:   :py:class:`int`
    :param back:     How many base pairs must be considered at the end of each edge
    :type back:      :py:class:`int`
    
    :returns: a string containing the fasta formated sequence
    :rtype: :py:class:`bytes`

    :raises: :py:class:`AssertionError`
    '''
    def getEdge(stemid):
        stem = assgraph.edgeIterator(edgePredicate=lambda e:assgraph.getEdgeAttr(e[0],e[1],e[2])['stemid']==stemid).next()
        return stem
        
        
    coverage=[]
    slength=[]
    seq=[]
    label=[]
    oldstem=None
    oldid=None
    
    for stemid in path:
        stem = getEdge(stemid)
        attr = assgraph.getEdgeAttr(*stem)

        if oldstem is not None:
            connected = pairEndedConnected(self,assgraph,oldid,stemid,back)
            
            if oldstem[1]==stem[0]:
                print >>sys.stderr,"Both segments %d and %d are connected" % (oldid,stemid)
                label.append('{connection: %d}' % connected)
            else:
                print >>sys.stderr,"Both segments %d and %d are disconnected" % (oldid,stemid)
                if connected > minlink:
                    print >>sys.stderr,"But linked by %d pair ended links" % connected
                    seq.append('N'*nlength)
                    label.append('{N-connection: %d}' % connected)
                else:
                    raise AssertionError('Disconnected path between stem '
                                         '%d and %d only %d pair ended links' % (oldid,stemid,connected))
        
        seq.append(attr['sequence'])
        coverage.append(attr['weight'])
        slength.append(attr['length'])
        label.append(attr['label'])
        
        oldstem = stem
        oldid=stemid
        
    s1 = getEdge(path[-1])
    s2 = getEdge(path[0])
    sid1=assgraph.getEdgeAttr(*s1)['stemid']
    sid2=assgraph.getEdgeAttr(*s2)['stemid']
    connected = pairEndedConnected(self,assgraph,sid1,sid2,back)
    
    if s1[1]==s2[0]:
        print >>sys.stderr,"Path is circular"
        circular=True
        label.append('{connection: %d}' % connected)
    else:
        if connected >= minlink:
            print >>sys.stderr,"Path is circular but disconnected" 
            print >>sys.stderr,"Linked by %d pair ended links" % connected
            label.append('{N-connection: %d}' % connected)
            circular=True
        else:
            print >>sys.stderr,"Path is linear" 
            circular=False
   
        
    sequence = ''.join(seq)
    length = sum(slength)
    mcov = sum(a * b for a,b in imap(None,slength,coverage)) / float(length)
    
    fasta=[">%s seq_length=%d; coverage=%5.1f; circular=%s; %s" % (identifier,length,
                                                                  mcov,bytes(circular),
                                                                  '.'.join(label))]
    fasta.extend(sequence[i:i+60] for i in xrange(0,len(sequence),60))
    return "\n".join(fasta)
    

